/*! \file std.h
  \brief My base file: contains a bunch of standard includes, macros, operators,
  etc. that are used by most other headers */
#ifndef MT_std_h
#define MT_std_h

//--------- standard includes:
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <math.h>
#include <vector>
#include <typeinfo>
#if defined MT_MSVC || defined MT_Windows || defined WIN32 || defined _WIN32 || defined _MSC_VER
#  undef  MT_MSVC
#  define MT_MSVC
#  include<limits>
#  include<time.h>
#  include<sys/timeb.h> 
#  ifdef MT_QT
#    include<windows.h>
#    undef min //I hate it that windows defines these macros!
#    undef max
#    undef  NOUNICODE
#    define NOUNICODE
//#    define QT_DLL
//#    define _MSC_VER 1300
//#    define _GDI32_
//#    define _MBCS
//#    define QT_THREAD_SUPPORT
#  endif
#  pragma warning(disable: 4305 4244 4250 4355 4786)
#else
#  include<unistd.h>
#  include<sys/time.h>
#  include<sys/times.h>
#  include<sys/resource.h>
#endif


//--------- my namespace:
/*! My main namespace */
namespace MT{
  inline int debugBreakpoint(){
    char c; c=0;
    //std::cout <<"HALT breakpoint..." <<std::endl; std::cin.getline(&c,1); //force to wait
    return 0; //insert a break point here, if you want...
  }
}


//--------- error handling:
#ifndef HALT
#  define MT_MSG(msg) { std::cerr <<__FILE__ <<':' <<__LINE__ <<": " <<msg <<std::endl; }
#  define HALT(msg) { std::cerr <<"HALT:" <<__FILE__ <<':' <<__LINE__\
                      <<":\n --- "<<msg<<" --- "<<std::endl; MT::debugBreakpoint(); abort();}
#endif


//--------- check macros:
#ifndef MT_NOCHECK
#  define CHECK(cond,msg) if(!(cond)) HALT("CHECK failed: "<<msg);
#else
#  define CHECK(cond,msg)
#endif


//--------- basic defs:
#define MT_PI 3.14159265358979323846
#define MT_2PI (2.*3.14159265358979323846)
#define MT_LN2 0.69314718055994528622676398299518041312694549560546875
typedef unsigned char byte; //!< byte
typedef unsigned int uint;  //!< unsigned integer
typedef void (*voidFct)();  //!< pointer to a void function


//--------- macros for looping
#ifndef forall
#define forall(i,a)            for(a.First(i)              ;a.CheckNext(i);    a.Next(i))
#define forall_rev(i,a)        for(a.Last(i)               ;a.CheckPrev(i);    a.Prev(i))
#define forall_save(i,j,a)     for(a.First(i),j=i,a.Next(j);a.CheckNext(i);i=j,a.Next(j))
#define forall_rev_save(i,j,a) for(a.Last(i) ,j=i,a.Prev(j);a.CheckPrev(i);i=j,a.Prev(j))
#endif


//--------- macros to define the standard << and >> operatos for most my classes:
#define stdInPipe(type)\
  inline std::istream& operator>>(std::istream& is,type& x){ x.read(is);return is; }
#define stdOutPipe(type)\
  inline std::ostream& operator<<(std::ostream& os,const type& x){ x.write(os); return os; }
#define stdPipes(type)\
  inline std::istream& operator>>(std::istream& is,type& x){ x.read(is);return is; }\
  inline std::ostream& operator<<(std::ostream& os,const type& x){ x.write(os); return os; }
#define fwdStdPipes(type)\
  inline std::istream& operator>>(std::istream& is,type& x);\
  inline std::ostream& operator<<(std::ostream& os,const type& x);


//--------- macros for piping doubles EXACTLY (without rounding errors) in hex coding:
#define OUTHEX(y) "0x" <<std::hex <<*((unsigned long*)&y) <<std::dec
#define INHEX(y) std::hex >>*((unsigned long*)&y) >>std::dec



//--------- standard conversion base type to a class:
/*! the basetype template `casts' an elemental type into a class */
template<class T>
class basetype{
public:
  T x; //!< the single datum
  basetype():x(){};
  basetype(const T _x):x(_x){};

  operator T() const{ return x; } //!< implicit and const access
  void read(std::istream& is){ is >>x; } //!< implements standard operator>>
  void write(std::ostream& os) const{ os <<x; } //!< implements standard operator<<
};
template<class T> stdInPipe(basetype<T>);
template<class T> stdOutPipe(basetype<T>);

namespace MT{
  typedef basetype<byte> Byte;
  typedef basetype<int> Int;
  typedef basetype<float> Float;
  typedef basetype<double> Double;
}


//--------- tools to scan (parse) from a stream"
namespace MT{
  inline char skip(std::istream&,char* skipchars=" \n\r\t"); //forward declaration..
}
/*! a global operator to scan (parse) strings from a stream */
inline std::istream& operator>>(std::istream& is,const char* str){
  uint i,n=strlen(str);
  char *buf=new char [n+1]; buf[n]=0;
  MT::skip(is," \n\r\t");
  is.read(buf,n);
  if(!is.good() || strcmp(str,buf)){
    for(i=n;i--;) is.putback(buf[i]);
    is.setstate(std::ios::failbit);
    std::cout <<"WARNING: scanning of constant string `" <<str
	      <<"' failed! (read instead: `" <<buf <<"')" <<std::endl;
  }
  delete[] buf;
  return is;
}
/*! the same global operator for non-const string */
inline std::istream& operator>>(std::istream& is,char* str){
  return is >>(const char*)str;
}



//===========================================================================
//
// what follows is somewhat messy organized:
// standard methods for my namespace
//

namespace MT{
  extern int argc;
  extern char** argv;
  extern std::ifstream cfgFile;
  extern bool cfgOpenFlag;
  extern std::vector<void*> parameters;
  //! stream modifier for some classes (Array in particular)
  extern bool IOraw;

  /*! access to the log-file */
  std::ofstream& log(const char *name="MT.log",bool _nolog=false);

  /*! open an output-file with name '#name#' */
  void open(std::ofstream& fs,const char *name,char* errmsg="");

  /*! open an input-file with name '#name#' */
  void open(std::ifstream& fs,const char *name,char* errmsg="");

  /*! a standard method to save an object into a file. The same as
    std::ofstream file; MT::open(file,filename); file <<x;
    file.close(); */
  template<class T>
  void save(const T& x,const char *filename){
    std::ofstream file;
    MT::open(file,filename);
    file <<x;
    file.close();
  }

  /*! a standard method to load object from a file. The same as
    std::ifstream file; MT::open(file,filename); file >>x;
    file.close(); */
  template<class T>
  void load(T& x,const char *filename){
    std::ifstream file;
    MT::open(file,filename);
    file >>x;
    file.close();
  }

  //! returns true if the (0-terminated) string s contains c 
  inline bool contains(const char* s,char c){
    for(uint i=0;s[i];i++) if(s[i]==c) return true;
    return false;
  }

  //! skips the chars (typically white characters) when parsing from the istream
  inline char skip(std::istream& is,char* skipchars){
    char c;
    do{ c=is.get(); } while(contains(skipchars,c));
    is.putback(c);
    return c;
  }

  //! skips a newline character (same as skip(is,"\n");)
  inline void skipLine(std::istream& is){
    char c;
    do{ c=is.get(); } while(c!='\n');
  }

  /*! returns the i-th of str */ 
  byte bit(byte *str,uint i);

  /*! flips the i-th bit of b */
  void flip(byte& b,uint i);

  /*! filps the i-th bit of b */
  void flip(int& b,uint i);

  /*! the distance between x and y w.r.t.\ a circular topology
      (e.g. modMetric(1,8,10)=3) */
  double modMetric(double x,double y,double mod);

  //! the sign (+/-1) of x (+1 for zero)
  double sign(double x);

  //! returns 0 for x<0, 1 for x>1, x for 0<x<1
  double linsig(double x);

  /*! the angle of the vector (dx,dy) in [-pi,pi] */
  float phi(float dx,float dy);

  /*! save division, checks for division by zero; force=true will return
    zero if y=0 */
  inline double DIV(double x,double y,bool force=false){
    if(x==0.) return 0.;
    if(force){ if(y==0.) return 0.; }else CHECK(y!=0,"Division by Zero!");
    return x/y;
  }

  inline double sigmoid11(double x){ return x/(1.+::fabs(x)); }

  template<class T> T MIN(T a,T b){ return a<b?a:b; }
  template<class T> T MAX(T a,T b){ return a>b?a:b; }

  //! approximate exp (sets up a static value table)
  double approxExp(double x);

  //! ordinary Log, but cutting off for small values
  double Log(double x);

  //! integer log2
  uint Log2(uint n);

  //! square of a double
  double sqr(double x);

  /*! real time since start of the process in floating-point seconds
    (probably in micro second resolution) -- Windows checked! */
  double realTime();

  /*! user CPU time of this process in floating-point seconds (pure
    processor time) -- Windows checked! */
  double cpuTime();

  /*! system CPU time of this process in floating-point seconds (the
    time spend for file input/output, x-server stuff, etc.)
    -- not implemented for Windows! */
  double sysTime();

  /*! total CPU time of this process in floating-point seconds (same
    as cpuTime + sysTime) -- not implemented for Windows! */
  double totalTime();
  
  /*! the absolute real time and date as string */
  char *date();

  /*! wait real time */
  void wait(double sec);

  /*! wait for an ENTER at the console */
  void wait();

  /*! the integral shared memory size -- not implemented for Windows! */
  long mem();

  /*! push reset on the stop watch (user CPU time) */
  void resetTimer();

  /*! read the stop watch (user CPU time) */
  double getTimer(bool reset=true);

  /*! memorize the command line arguments and open a log file */
  void init(int _argc, char *_argv[]);

  /*! returns false if command-line option #-tag# was not found */
  bool checkOption(const char *tag);

  /*! returns false if command-line option #-tag# was not found */
  bool getOption(const char *tag,char* &option);

  /*! Open a (possibly new) config file with name '#name#'.<br> If
    #name# is not specified, it searches for a command line-option
    '-cfg' and, if not found, it assumes #name=MT.cfg# */
  void openConfigFile(char* name=0);

#ifdef MT_QT
  /*! create a Qt application */
  void initQt();
  //! dummy struct which calls MT::initQt(); as constructor
  struct InitQt{ InitQt(){ initQt(); } };
#else
  inline void initQt(){ MT_MSG("Warning: initQt without QT configured"); }
#endif
}



//===========================================================================
//
// Parameter class - I use it frequently to read parameters from file or cmd line
// 

namespace MT{
  /*! A parameter that initializes itself from the command line,
    parameter file, or a default value (priority in this order).
    Initialization is done on the fly the _first_ time its value is
    queried (i.e., referenced by the cast operators).*/
  template<class type>
  class Parameter{
  public:
    const char* typeName;
    type value,Default;
    char* tag;
    bool initialized,hasDefault;


  public:
    //{ constructors

    /*! Determines the tag to search for in parameter file/command
      line */
    Parameter(char* _tag){
      typeName=typeid(type).name();
      MT::parameters.push_back(this);
      initialized=false;
      tag=_tag;
      hasDefault=false;
    };

    /*! specifies also a default value -- parameter does not have to but
      can be specified in the parameter file/command line */
    Parameter(char* _tag,const type& _default){
      typeName=typeid(type).name();
      MT::parameters.push_back(this);
      initialized=false;
      tag=_tag;
      hasDefault=true;
      Default=_default;
    };

    ~Parameter(){
      std::vector<void*>::iterator it;
      for(it=MT::parameters.begin();it!=MT::parameters.end();it++)
	if(*it==this){ MT::parameters.erase(it); return; }
    }

    //{ value access

    /*! standard type conversion: returns a const of the parameter value */
    operator type(){ if(!initialized) initialize(); return value; }

    /*! ()-operator: returns an lvalue of the parameter value */
    type& operator()(){ if(!initialized) initialize(); return value; }


    //{ manipulation

    /*! assigs a value to the parameter -- no further initialization needed */
    type& operator=(const type v){ initialized=true; value=v; return value; }

    /*! set the tag (replacing the one from the constructor) */
    void setTag(char* _tag){ tag=_tag; }

    /*! enforces that the parameter is reinitialized from the parameter
      file/command line, the next time it is referenced -- even if it
      has been initialized before */
    void reInitialize(){ initialized=false; }


    //{ explicit grabbing

    /*! Search for a command line option #-tag# and, if found, pipe the
      next command line option into #value# by the
      #operator>>(istream&,type&)#. Returns false on failure. */
    bool grabFromCmdLine(){
      char* opt;
      if(!getOption(tag,opt)) return false;
      std::istringstream s(opt);
      s >>value;
      if(s.fail()) HALT("error when reading parameter from command line: " <<tag);
      return true;
    }

    /*! Search the first occurence of a sequence '#tag#:'
      in the config file (opened automatically) and, if found, pipes
      it in #value#. Returns false if parameter is not found. */
    bool grabFromCfgFile(){
      if(!cfgOpenFlag){
	log() <<"..." <<std::endl;
	openConfigFile();
	log() <<std::setw(40) <<"..." <<": " <<std::setw(5);
      }
      cfgFile.clear();
      cfgFile.seekg(std::ios::beg);
      if(!cfgFile.good()) return false;

      unsigned n=strlen(tag);
      char *buf=new char [n+2]; memset(buf,0,n+2);
      while(cfgFile.good()){
	memmove(buf,buf+1,n);
	buf[n]=cfgFile.get();
	if(buf[n]==' ' || buf[n]==':' || buf[n]=='='){ buf[n]=0; if(!strcmp(tag,buf)) break; buf[n]=':'; }
      };
      delete[] buf;

      if(!cfgFile.good()) return false;
 
      skip(cfgFile," :=\n\r\t");
      cfgFile >>value;
 
      if(cfgFile.fail()) HALT("error when reading parameter " <<tag);
      return true;
    }
    //}

  private:
    void initialize(){
      if(!initialized){
        if(!cfgOpenFlag) openConfigFile();
	initialized=1;
	if(!tag) HALT("uninitialized parameter without tag!");
	log() <<std::setw(20) <<tag <<": " <<std::setw(5);
	log().flush();

	if(grabFromCmdLine()){
	  log() <<value <<" [" <<typeid(value).name() <<"] (cmd line!)" <<std::endl;
	  return;
	}

	if(grabFromCfgFile()){
	  log() <<value <<" [" <<typeid(value).name() <<"]" <<std::endl;
	  return;
	}

	if(hasDefault){
	  value=Default;
	  log() <<value <<" [" <<typeid(value).name() <<"] (default!)"<<std::endl;
	  return;
	}

	HALT("could not initialize parameter `" <<tag
	     <<"': parameter has no default;\n     either use command option `-"
	     <<tag <<" ...' or specify `"
	     <<tag <<"= ...' in the config file `MT.cfg'");
      }
    }
  };

}



//===========================================================================
//
// random number generator
//

namespace MT{
  /*! A random number generator. An instantiation #rnd# of a #Rnd#
    object is created. Use this one object to get random numbers.*/
  class Rnd{
  private:
    bool ready;
    int rpoint;     /* Feldindex    */
    long rfield[256];   /* Schieberegisterfeld  */


  public:
    //! ...
    Rnd(){ ready=false; };


  public://{ Initialization
    /*! initialize with a specific seed */
    unsigned long seed(unsigned long n);

    /*! use Parameter<uint>("seed") as seed */
    unsigned long seed();

    /*! uses the internal clock to generate a seed */
    unsigned long clockSeed();

  public://{ Access
    /*! a initeger random number uniformly distributed in [0,?] */
    inline uint num(){ if(!ready) seed(); return (uint)rnd250() >> 5; }
    /*! same as #num()# */
    inline uint operator ()(){ return num(); }
    /*! a initeger random number uniformly distributed in [0,#i#-1] */
    inline uint num(uint limit){
      CHECK(limit,"zero limit in rnd.num()"); return num() % limit; }
    inline uint num(int lo,int hi){ return lo+num(hi-lo+1); }
    /*! same as #num(i)# */
    inline uint operator ()(uint i){ return num(i); }
    inline uint operator ()(int lo,int hi){ return num(lo,hi); }
    /*! a random variable uniformly distributed in [0,1] */
    inline double uni(){ return ((double)num(1<<22))/(1<<22); }
    /*! a random variable uniformly distributed in [#low#, #high#] */
    inline double uni(double low,double high){ return low+uni()*(high-low); }
    /*! a gaussian random variable with mean zero */
    double gauss(double stdDev);
    /*! a positive integer drawn from a poisson distribution with given
      #mean#; is case #mean#>100, a (positive) gauss number
      #floor(mean+gauss(sqrt(mean))+.5)# is returned */
    uint poisson(double mean);
    //}


  private:
    inline long  rnd250(){
      rpoint = (rpoint+1) & 255;          // Index erhoehen
      return rfield[rpoint] =  rfield[(rpoint-250) & 255]
	                     ^ rfield[(rpoint-103) & 255];
    }

    void seed250(long seed);
  };

  //! The global Rnd object
  extern Rnd rnd;
}



//===========================================================================
//
// string class
//

namespace MT{
  typedef std::iostream IOStream;
  
  /*! String implements the functionalities of an ostream and an
  istream, but also can be send to an ostream or read from an
  istream.
  
    It is based on a simple streambuf derived from the std::vector<char>
  class */
  class String:public IOStream{
  private:
    class StringBuf:public std::streambuf{
    public:
      std::vector<char> arr;
      const char* ptr() const{ return &(arr[0]); }
      char* ptr(){ return &(arr[0]); }
      virtual int overflow(int C = EOF){
	arr[arr.size()-1]=C; arr.push_back(0); resetI();
	return C;
      }
      void resetI(){ setg(ptr(),ptr(),ptr()+arr.size()-1); }
      char *getIpos(){ return gptr(); }
      void setIpos(char* p){
	setg(ptr(),p,ptr()+arr.size()-1);
      }
    };
    
    public:
      StringBuf buf;
      //! standard constructor
      String():IOStream(&buf){ clr(); }
      //! copy constructor
      String(const String& s):IOStream(&buf){ this->operator=(s); }
      //! copy constructor for an ordinary C-string (needs to be 0-terminated)
      String(const char* s):IOStream(&buf){ this->operator=(s); }
      //! returns a reference to this
      IOStream& str(){ return (IOStream&)(*this); }
      /*! returns the true memory buffer (C-string) of this class (which is
      always 0-terminated) */
      operator char*(){ return buf.ptr(); }
      char* ptr(){ return buf.ptr(); }
      char operator()(uint i){ return buf.arr[i]; }
      //! as above but const
      operator const char*() const{ return buf.ptr(); }
      //! copy operator
      void operator=(const String& s){ buf.arr=s.buf.arr; buf.resetI(); }
      //! copies from the C-string
      void operator=(const char *s){ buf.arr.resize(strlen(s)+1); memmove(buf.ptr(),s,strlen(s)+1); buf.resetI(); }
      //! shorthand for the !strcmp command
      bool operator==(const char *s){ return !strcmp(buf.ptr(),s); }
      //! shorthand for the !strcmp command
      bool operator==(const String& s){ return !strcmp(buf.ptr(),s.buf.ptr()); }
      //! deletes all memory and resets all stream flags
      String& clr(){ buf.arr.resize(1); buf.arr[0]=0; clear(); return *this; }
      /*! when using this String as an istream (to read other variables
      from it), this method resets the reading-pointer to the beginning
      of the string and also clears all flags of the stream */
      String& resetI(){ buf.resetI(); clear(); return *this; }
      char *getIpos(){ return buf.getIpos(); }
      void setIpos(char* p){buf.setIpos(p); }
      
      //! returns the length of the string (excluding the terminating 0)
      uint N() const{ return buf.arr.size()-1; }
      /*! appends some other text/variable/whatever to the string (as if
      piped into it) */
      template<class T>
      String operator+(const T& v) const{
	String news(*this); news <<v; return news;
      }
      //! writes the string into some ostream
      void write(std::ostream& os) const{ os <<buf.ptr(); }
      /*! reads the string from some istream until one of the stopSymbols
      is encountered (default: newline symbols) */
      void read(std::istream& is,char *stopSymbols="\n\r"){
	buf.arr.resize(0);
	char c=is.get();
	while(is.good() && !MT::contains(stopSymbols,c)){
	  buf.arr.push_back(c); c=is.get(); }
	is.putback(c);
	is.clear();
	buf.arr.push_back(0);
	resetI();
      }
  };
}
stdPipes(MT::String);

#define STRING(x) ((MT::String&)(MT::String().str()<<x)).ptr()



//===========================================================================
//
// color class
//

namespace MT{
  /*! simple float[3] color class */
  class Color{
  public:
    float
      r, //!< red
      g, //!< green
      b; //!< blue
    
    //! ...
    friend inline Color operator+(const Color& c1,const Color& c2){
      return Color(c1.r+c2.r, c1.g+c2.g, c1.b+c2.b); }
    
    //! ...
    friend inline Color operator*(float f,const Color& c2){
      return Color(f*c2.r, f*c2.g, f*c2.b); }
    
  public:
    //! initializes to white
    Color(){ setGray(1.); }
    
    //! initialize with RGB
    Color(float red,float green,float blue){ setRgb(red,green,blue); }
    
    //! copy operator
    Color& operator=(const Color& c){ r=c.r; g=c.g; b=c.b; return *this; }
    
    //! return true iff black
    bool operator!(){ if(r==0. && g==0. && b==0.) return true; return false; }
    
    //! float-pointer access
    operator const float*() const{ return (float*)this; }
    
    //! chooses color from a color table (distributed around the hue-scale)
    void setIndex(unsigned i){
      if(!i) setRgb(0.,0.,0.); else setHsv(((i-1)*63)%360,255,255); }
    
    //! set RGA values
    void setRgb(float red,float green,float blue){ r=red; g=green; b=blue; }
    
    //! set RGA values from bytes in [0,255]
    void setRgbByte(byte red,byte green,byte blue){
      r=(float)red/255.; g=(float)green/255.; b=(float)blue/255.; }
    
    //! set color by hue [0,360], saturation [0,255], and value [0,255]
    void setHsv(int hue,byte sat,byte val){
      float h=(double)hue/60.,s=(double)sat/255.,v=(double)val/255.;
      h=fmod(h,6);
      r=g=b=0.;
      if(h<=1.)        { r=v; g=v*h; }
      if(h>1. && h<=2.){ g=v; r=v*(2.-h); }
      if(h>2. && h<=3.){ g=v; b=v*(h-2.); }
      if(h>3. && h<=4.){ b=v; g=v*(4.-h); }
      if(h>4. && h<=5.){ b=v; r=v*(h-4.); }
      if(h>5. && h<=6.){ r=v; b=v*(6.-h); }
      r=s*r+(1-s)*v;
      g=s*g+(1-s)*v;
      b=s*b+(1-s)*v;
    }
    
    //! set color by temperature: hot=red, middle=yellow, cold=blue
    void setTemp(float temp){
      Color hot(1.,0.,0.),middle(1.,1.,0.),cold(0.,0.,1.);
      if(temp>1.) temp=1.;
      if(temp<0.) temp=0.;
      if(temp>.5){ temp=2.*temp-1.; *this=temp*hot + (1.-temp)*middle; }
      else{ temp=2.*temp; *this=temp*middle + (1.-temp)*cold; }
    }
    
    //! set color by temperature: red - yellow - gray(middle) - green - blue
    void setTemp2(float temp){
      Color r(1.,0.,0.),y(1.,1.,0.),zero(.5,.5,.5),g(0.,1.,0.),b(0.,0.,1.);
      if(temp>1.) temp=1.;
      if(temp<-1.) temp=-1.;
      if(temp>.5){  temp=2.*temp-1.; *this=temp*r + (1.-temp)*y; return; }
      if(temp>.0){  temp=2.*temp;    *this=temp*y + (1.-temp)*zero; return; }
      if(temp>-.5){ temp=-2.*temp;   *this=temp*g + (1.-temp)*zero; return; }
      { temp=-2.*temp-1.;*this=temp*b + (1.-temp)*g; return; }
    }
    
    //! set gray value [0,1]
    void setGray(float gray){ if(gray<0) gray=0.; if(gray>1) gray=1.; r=g=b=gray; }
    
    //! get RGB values as bytes [0,255]
    void getRgb(byte& R,byte& G,byte& B) const{
      R=(byte)(255.*r); G=(byte)(255.*g); B=(byte)(255.*b); }
    
    //! get the gray value (average of RGB)
    float getGray() const{ return (r+g+b)/3.; }
    
    //! mix with white
    void whiten(float f){
      if(f>1.) f=1.; else if(f<0.) f=0.;
      r+=f*(1.-r); g+=f*(1.-g); b+=f*(1.-b);
    }
    
    //! mix with black
    void blacken(float f){
      if(f>1.) f=1.; else if(f<0.) f=0.;
      r-=f*r; g-=f*g; b-=f*b;
    }
    
    //! plain color mixing
    void mix(Color& A,Color& B,float f=.5){
      if(f>1.) f=1.; else if(f<0.) f=0.;
      r=f*A.r+(1.-f)*B.r;
      g=f*A.g+(1.-f)*B.g;
      b=f*A.b+(1.-f)*B.b;
    }
    
    //! additive color mixing
    void mixAdd(Color& A,Color& B,float f=.5){
      if(f>1.) f=1.; else if(f<0.) f=0.;
      r=1.-f*(1.-A.r)+(1.-f)*(1.-B.r);
      g=1.-f*(1.-A.g)+(1.-f)*(1.-B.g);
      b=1.-f*(1.-A.b)+(1.-f)*(1.-B.b);
    }
    
    //! subtractive color mixing
    void mixSub(Color& A,Color& B,float f=.5){
      if(f>1.) f=1.; else if(f<0.) f=0.;
      r=1.-::pow(1.f-A.r,f)*::pow(1.f-B.r,1.f-f);
      g=1.-::pow(1.f-A.g,f)*::pow(1.f-B.g,1.f-f);
      b=1.-::pow(1.f-A.b,f)*::pow(1.f-B.b,1.f-f);
    }
    
    //! take smaller of the two values
    void min(Color& A,Color& B){
      r=A.r<B.r?A.r:B.r;
      g=A.g<B.g?A.g:B.g;
      b=A.b<B.b?A.b:B.b;
    }

    //! prototype for operator <<
    void write(std::ostream& os) const{ os <<"(" <<r <<":" <<g <<":" <<b <<")"; }
    
    //! prototype for operator >>
    void read(std::istream& is){ is >>"(" >>r >>":" >>g >>":" >>b >>")"; }
  };
}
stdPipes(MT::Color);


//===========================================================================
//
// gnuplot calls
//

void gnuplot(const char* command);
void gnuplotEPS(const char* file,const char* command);


//===========================================================================
//
// implementations
//

#ifdef MT_IMPLEMENTATION
#  include"std.cpp"
#endif



//===========================================================================
//
// USING's
//

//using std::cout;
//using std::endl;
using std::ofstream;
using std::ifstream;
using MT::rnd;
using MT::Color;
using MT::Parameter;
using MT::String;


#endif
