#include "lwpr.h"

//===========================================================================
//
// helper methods
//

doubleA matabs(const doubleA& a){
  doubleA b; b.resize(a); uint i; for(i=0;i<a.N;i++) b.p[i]=fabs(a.p[i]); return b;
}



//===========================================================================
//
// the local models (``receptive fields'') of LWPR
//

void Lmodel::loadParameters(uint N){
  static Parameter<double> p_initD("initD",0.1);
  static Parameter<double> p_init_lambda("init_lambda",0.999);
  static Parameter<double> p_tau_lambda("tau_lambda",0.99999);
  static Parameter<double> p_final_lambda("final_lambda",0.9999);
  static Parameter<double> p_add_threshold("add_threshold",0.5);
  static Parameter<bool> p_meta("meta",true);
  static Parameter<double> p_meta_rate("meta_rate",0.5);
  static Parameter<int> p_r("R",2);
  static Parameter<double> p_alpha("alpha",0.01);
  static Parameter<double> p_gamma("gamma",1e-6);

  initD=p_initD;
  init_lambda = p_init_lambda;
  tau_lambda = p_tau_lambda;
  final_lambda= p_final_lambda;
  add_threshold = p_add_threshold;
  meta= p_meta;
  meta_rate = p_meta_rate;
  R = p_r;
  alpha.resize(N,N);
  alpha = p_alpha;
  gamma = p_gamma/N;
}

/**
* Constructs a local model for input dimensions
* equal to N
*/
Lmodel::Lmodel(const doubleA& x, double y){
  int N = x.d0;//number of rows in x. Equivalent to input dims

  loadParameters(N);

  mean_x.resize(N);
  mean_x = 0;
  mean_y = 0;
  
  apk = 0;
  
  center=x;
  dist.resize(N,N);
  //initialises to identity matrix
  int i,j;
  for(i=0;i<N;i++) for(j=0;j<N;j++) dist(i,j) = (i==j)?initD:0;
  
  cholesky(M,dist);
  
  sum_e2 = 0;
  
  b.resize(N,N);
  b = log(alpha + 1e-10);
  h.resize(N,N);
  h = 0;
  
  initialise(N);
  //XXX : don't know why y shld be 
  //updated while others ignored
  //mean_y = y; @todo REMOVED!
  betas(0) = mean_y;
  
  trustworthy = false;
  degenerated=false;
  //updateMean(x,y);
  //update(x,y);
}

/**
* Copy constructor
**/
Lmodel::Lmodel(const Lmodel& model, const doubleA& x, double y){
  int N = x.d0;//number of rows in x

  loadParameters(N);
  
  mean_x.resize(N);
  mean_x = 0;
  mean_y = 0;
  
  apk = 0;
  
  
  center=x;
  sum_e2 = 0;
  
  
  h.resize(N,N);
  h = 0;
  dist = model.dist;
  M = model.M;
  alpha = model.alpha;
  mean_y = model.mean_y;
  b.resize(N,N);
  b = log(alpha + 1e-10);
  
  initialise(N);
  betas(0) = mean_y;
  
  degenerated=false;
  trustworthy = false;
}

/*void Lmodel::reinitialize(){
  R--;
  static Parameter<double> p_alpha("alpha",0.01);
  
  int N = mean_x.d0;//number of rows in x
  //alpha.resize(N,N);
  alpha = p_alpha;
  static Parameter<double> p_gamma("gamma",1e-6);
  gamma = p_gamma/N;
  mean_x = 0;
  //mean_y = 0;
  //center = new doubleA (x);
  //dist.resize(N,N);
  //initialises to identity matrix
  int i,j;
  for(i=0;i<N;i++) for(j=0;j<N;j++) dist(i,j) = (i==j)?initD:0;
    
    cholesky(M,dist);
    
    sum_e2 = 0;
    
    //b.resize(N,N);
    b = log(alpha + 1e-10);
    //h.resize(N,N);
    h = 0;
    
    initialise(N);
    //XXX : don't know why y shld be 
    //updated while others ignored
    //mean_y = y;
    betas(0) = mean_y;
    
    trustworthy = false;
    degenerated=false;
    //updateMean(x,y);
    //update(x,y);
    
}*/

void Lmodel::initialise(int dim){
  uprojections.resize(R);
  pprojections.resize(R);
  sXresYres.resize(R);
  betas.resize(R+1);
  axz.resize(R);
  azz.resize(R);
  azres.resize(R);
  MSE.resize(R);
  lambda.resize(R);
  n_data.resize(R);
  aH.resize(R);
  aG.resize(R);
  W.resize(R);
  MyMSE_R = 0;
  sum_ecv2.resize(R);
  aE = 0;
  aH = 0;
  aG = 0;
  W = 1e-10;
  sum_ecv2 = 0;
  
  int r,i;
  for(r=0;r<R;r++){
    uprojections(r).resize(dim);	
    pprojections(r).resize(dim);
    sXresYres(r).resize(dim);
    axz(r).resize(dim);
    //make them an identity matrix
    for(i=0;i<dim;i++){
      uprojections(r)(i) = (i == r) ? 1:0;
      pprojections(r)(i) = (i == r) ? 1:0;
      axz(r)(i) = 0;
    }
    sXresYres(r)=0.0;
    betas(r) = 0;
    azz(r) = 1e-10;
    azres(r) = 0;
    MSE(r) = 0;
    lambda(r) = init_lambda;
    n_data(r) = 1e-10;
  }
  betas(R) = 0;
}

/**
* Computes the activation for a given training
* data pair x
*/
double Lmodel::activation(const doubleA& x){
  static doubleA d;
  MT::minus(d,x,center);
  double ex1=scalarProduct(dist,d,d);
  double w = exp(-0.5 * ex1);	
  if(w != w){ //testing for nan
    HALT("don't go here!");
    /*
    std::cout << "Activation too large, reinitializing local model" << std::endl;
    //std::cout<<center<<std::endl;
    reinitialize();
    degenerated=true;		
    //std::cout << "diff " << diff<<std::endl;
    //std::cout << "dist " << dist<<std::endl;
    //std::cout << "ex " << ex<<std::endl;
    //std::cout << "ex1 " << ex1<<std::endl;
    
    return activation(x);
    */
  }
  return w;
}

/**
* Update the mean and stuff
*/
void Lmodel::updateMean(const doubleA& x, double y){
  double w = activation(x);
  double newW = lambda(0)*W(0) + w;
  mean_x = (lambda(0)*W(0)*mean_x + w*x)/newW;
  mean_y = (lambda(0)*W(0)*mean_y + w*y)/newW; // equivalent to `beta0' in paper
}

/**
* Compute the current prediction error
*/
void Lmodel::updateError(const doubleA& x, double y){
  //residuals
  doubleA xres(x - mean_x);
  double yres = mean_y;
  
  //int nreg = uprojections.N;
  
  doubleA zs; computeZs(zs,x);
  double w = activation(x);
  
  for(int r=0;r<R;r++){
    double beta = betas(r+1);
    double z = zs(r);
    yres += beta*z;
    MSE(r) = lambda(r)*MSE(r) + w*pow((y - yres),2); // Compare this with MATLAB!
  }	
}

void Lmodel::updateError_matlab(double e_cv, double w){
  // I set this to zero whenever I add a new projection R
  MyMSE_R = lambda(0)*MyMSE_R + w*e_cv*e_cv;
}

void Lmodel::computeZs(doubleA& zs,const doubleA& x){
  doubleA xres(x - mean_x);

  zs.resize(R);
  
  for(int r=0;r<R;r++){
    double z = scalarProduct(xres,uprojections(r));
    zs(r) = z;
    xres -= z*pprojections(r);
  }
}

/**
* Update the local regression and compute residuals
*/
void Lmodel::update(const doubleA& x, double y){
  double res = y - mean_y;
  static doubleA xres;
  MT::minus(xres,x,mean_x);

  doubleA zs; computeZs(zs,x);
  double w = activation(x);
  
  betas(0) = mean_y;
  doubleA lambda_slow(1.-(1.-lambda)/10.);

  int r;
  for(r=0;r<R;r++){
    double z = zs(r);
    azz(r) = lambda(r)*azz(r) + w*pow(z,2);
    azres(r) = lambda(r)*azres(r) + w*z*res;
    
    axz(r) = lambda(r)*axz(r) + w*xres*z;
    
    //update projections
    //doubleA u(lambda_slow(r)*uprojections(r) + w*xres*res);
    sXresYres(r)=sXresYres(r)*lambda_slow(r)+w*xres*res;
    //uprojections(r) = u/sqrt(matmult(~u,u)(0));
    double norm_coef=norm(sXresYres(r));
    //uprojections(r) = sXresYres(r)/sqrt(matmult(~sXresYres(r),sXresYres(r))(0));
    if(norm_coef!=0) uprojections(r) =sXresYres(r)/norm_coef;
    xres -= z*pprojections(r);
    pprojections(r) = axz(r)/azz(r);
    res -= z*betas(r+1);
    betas(r+1) = azres(r)/azz(r);
  }
  
  int n_in = x.d0;
  
  trustworthy = true;
  for(r=0;r<R;r++) 
    trustworthy = trustworthy && (n_data(r) > 2*n_in);
}

/**
* Update simple stats
**/
void Lmodel::updateStat(double w){
  W = lambda%W + w;
  n_data = lambda%n_data + 1.;
  lambda = tau_lambda * lambda + final_lambda*(1. - tau_lambda);
}

/**
* Predicts the value of y given x
*/
double Lmodel::predict(const doubleA& x) const {
  double yq = betas(0);
  doubleA xq(x - mean_x);
  
  for(int r=0;r<R;r++){
    double s = scalarProduct(xq,uprojections(r));
    yq += betas(r+1)*s;
    xq -= s * pprojections(r);
  }
  
  return yq;
}

/**
* gets e for a given x and y
**/
double Lmodel::gete(const doubleA& x, double y){
  double pred = predict(x);
  return (y-pred);
}

/**
* gets e_cv for a given x and y
**/
double Lmodel::gete_cv(const doubleA& x, double y){
  double pred = predict(x);
  return (y-mean_y-pred+betas(0));
}

/**
* Checks if a projection needs to be added
**/
bool Lmodel::check_add_projection(const doubleA& x, double y){
  int n_reg = uprojections.N;
  int n_in = x.d0;
  
  return (!degenerated) && (n_reg<n_in) && 
    (MSE(n_reg-1)/MSE(n_reg-2)<add_threshold) &&
    n_data(n_reg-1)/n_data(0) > 0.99 &&
    n_data(n_reg-1)*(1-lambda(n_reg-1)) > 0.5;	
}

/**
* Adds a new projection
**/
void Lmodel::add_projection(const doubleA& x, double y){
  std::cout << "Adding a projection" << std::endl;
  //std::cout<<center<<std::endl;
  int dim = x.d0;
  
  betas.append(0);
  (uprojections).append(doubleA (dim));	
  (pprojections).append(doubleA (dim));	
  (sXresYres).append(doubleA (dim));	
  (axz).append(doubleA (dim));
  //make them an identity matrix
  for(int i=0;i<dim;i++){
    uprojections(R)(i) = (i == R) ? 1:0;
    pprojections(R)(i) = (i == R) ? 1:0;
    axz(R)(i) = 0;
    sXresYres(R)(i)=0.0;
  }
  azz.append(1e-10);
  azres.append(0);
  MSE.append(0);
  lambda.resizeCopy(R+1);
  lambda(R) = init_lambda;
  n_data.resizeCopy(R+1);
  n_data(R) = 0;
  aH.resizeCopy(R+1);
  aH(R) = 0;
  aG.resizeCopy(R+1);
  aG(R) = 0;
  W.resizeCopy(R+1);
  W(R) = 1e-10;
  MyMSE_R = 0;
  sum_ecv2.resizeCopy(R+1);
  sum_ecv2(R) = 0;
  R++;
}

/**
* Evaluate the update of M - dJdM
**/
void Lmodel::getdJdM(doubleA& dJdM,const doubleA& D, const doubleA& M,
				const doubleA& x, const doubleA& center,
				const doubleA& z, const doubleA& azz, double w,
				double W, double e_cv, double e, double gamma,
				const doubleA& derivative_ok, double& sum_dJ1dw,
				doubleA& dwdM){
  doubleA dJ2dM;
  sum_dJ1dw = getsum_dJ1dw(e_cv, e, w, W, z, azz, derivative_ok);
  //std::cout<<"sum_dJ1_dw "<<*sum_dJ1dw <<std::endl;
  getdwdM(dwdM,M,x,center,w);
  //std::cout<<"dwdM "<<*dwdM<<std::endl;
  getdJ2dM(dJ2dM,D,M,gamma);
  //std::cout<<"dJ2dM "<<dJ2dM<<std::endl;
  //std::cout<<"w/W "<<w/W<<std::endl;
  dJdM = sum_dJ1dw*dwdM + w/W * dJ2dM;
  //doubleA tmp(sum_dJ1dw*dwdM);
  //doubleA tmp1(w/W * dJ2dM);
  //std::cout<<"dJdM "<<dJdM<<std::endl;
  //return dJdM;
}

/**
* Evaluate the value of sum_dJ1dw 
**/
double Lmodel::getsum_dJ1dw(double e_cv, double e, double w, double W,
			    const doubleA& z, const doubleA& azz, const doubleA& derivative_ok){
  doubleA z2(z%z);
  doubleA q(z%derivative_ok/azz);
  doubleA q2(q%q);

  double qTaH = scalarProduct(q,aH); //((doubleA ) matmult(~q,aH))(0);
  double q2TaG = scalarProduct(q2,aG); //((doubleA ) matmult(~q2,aG))(0);
  aE = lambda(0)*aE + w*e_cv*e_cv;
  /*std::cout<<"e_cv "<<e_cv<<std::endl;
  std::cout<<"W "<<W<<std::endl;
  std::cout<<"e "<<e<<std::endl;
  std::cout<<"q "<<q<<std::endl;
  std::cout<<"q2 "<<q2<<std::endl;
  std::cout<<"z "<<z<<std::endl;
  std::cout<<"z2 "<<z2<<std::endl;
  std::cout<<"azz "<<azz<<std::endl;
  std::cout<<"qTaH "<<qTaH<<std::endl;
  std::cout<<"q2TaG "<<q2TaG<<std::endl;
  std::cout<<"aE "<<aE<<std::endl;
  std::cout<<"aH "<<aH<<std::endl;
  std::cout<<"aG "<<aG<<std::endl;
  */
  double sum_dJ1dw = (e_cv*e_cv - 2*e*qTaH - 2*q2TaG)/W - aE/(W*W);
  //std::cout<<"sum_dJ1dw "<<sum_dJ1dw<<std::endl;
  return sum_dJ1dw;
}

/**
* Update the statistics associated with the calculation for the update
* of D
**/
void Lmodel::update_dist_stat(double e_cv, double w, const doubleA& z,
			      const doubleA& azz,double transient_multiplier,
			      const doubleA& derivative_ok){
  doubleA z2(z%z);
  doubleA q(z%derivative_ok/azz);
  //std::cout<<std::endl<<"derivative_ok "<<derivative_ok<<std::endl;
  //std::cout<<"transient_multiplier "<<transient_multiplier<<std::endl;
  //std::cout<<"w "<<w<<std::endl;
  double h = w*scalarProduct(z,q); //(((doubleA ) matmult(~z,q))(0));
  //std::cout<<"z "<<z<<std::endl;
  //std::cout<<"q "<<q<<std::endl;
  //std::cout<<"h "<<h<<std::endl;
  //std::cout<<"e_cv "<<e_cv<<std::endl;
  //std::cout<<"p aH "<<aH<<std::endl;
  doubleA H(lambda(0)*aH + w*e_cv*z*transient_multiplier/(1.-h));
  aH = derivative_ok%H + (1.-derivative_ok)%aH;
  //std::cout<<"m aH "<<aH<<std::endl;
  //std::cout<<"p aG "<<aG<<std::endl;
  //std::cout<<"aH "<<aH<<std::endl;
  //std::cout<<"w2 "<<w*w<<std::endl;
  //std::cout<<"ecv2 "<<e_cv*e_cv<<std::endl;
  //std::cout<<"z2 "<<z2<<std::endl;
  doubleA G(lambda(0)*aG + w*w*e_cv*e_cv*z2*transient_multiplier/(1.-h));
  aG = derivative_ok%G + (1.-derivative_ok)%aG;
  //std::cout<<"m aG "<<aG<<std::endl;
}

/**
* Evaluate the value of dJ2/dM
**/
void Lmodel::getdJ2dM(doubleA& dJ2dM,const doubleA& D, const doubleA& M, 
				 double gamma){
  dJ2dM.resize(M);
  int i,j,k,l,size=M.d0;
  dJ2dM = 0;
  doubleA dDdMkl;
  double sum;

  for(k=0;k<size;k++) for(l=k;l<size;l++){
    sum=0.;
    getdDdMkl(dDdMkl,M,k,l);
    for(i=0;i<size;i++) for(j=0;j<size;j++)
      sum += D(i,j)*dDdMkl(i,j);
    dJ2dM(k,l) = 2*gamma*sum; //gamma has already been divided by size
    
  }
}

/**
* Evaluate the value of dJ2dJ2/dMdM
**/
void Lmodel::getdJ2dJ2dMdM(doubleA& dJ2dJ2dMdM,const doubleA& D, const doubleA& M, 
				      double gamma){
  dJ2dJ2dMdM.resize(M);
  int i,j,k,l,size = M.d0;
  dJ2dJ2dMdM = 0;
  doubleA dDdMkl;
  
  for(k=0;k<size;k++) for(l=k;l<size;l++){
    double sum = 0;	
    getdDdMkl(dDdMkl,M,k,l);
    for(i=0;i<size;i++) for(j=0;j<size;j++)
      //XXX there is an extra term of 2 in matlab
      sum += pow(dDdMkl(i,j),2);
    
    //XXX what's special about k,k
    dJ2dJ2dMdM(k,l) = 2*gamma*(2*D(k,k) + sum); //gamma has already been divided by size
    dJ2dJ2dMdM(k,l) = dJ2dJ2dMdM(l,k);
  }
}

/**
* Evaluate the value of dw/dM
*/
void Lmodel::getdwdM(doubleA& dwdM,const doubleA& M, const doubleA& x, 
				const doubleA& center, double w){
  dwdM.resize(M);
  int k,l,size = M.d0;
  dwdM = 0;
  doubleA dDdMkl;
  
  for(k=0;k<size;k++) for(l=k;l<size;l++){
    getdDdMkl(dDdMkl,M,k,l);
    dwdM(k,l) = -0.5*w*sqrDistance(dDdMkl,x,center);
  }
  //std::cout << "dwdM " << dwdM << std::endl;
}

/**
* Evaluate the value of dwdw/dMdM
*/
void Lmodel::getdwdwdMdM(doubleA& dwdwdMdM,const doubleA& M, const doubleA& x, 
				    const doubleA& center, double w){
  doubleA dwdM;
  getdwdM(dwdM,M,x,center,w);
  int k,l,size = M.d0;
  dwdwdMdM.resize(M);
  dwdwdMdM = 0;
  
  for(k=0;k<size;k++) for(l=k;l<size;l++){
    doubleA tmp(x-center);
    dwdwdMdM(k,l) = dwdM(k,l)*dwdM(k,l)/w - w*pow(tmp(l),2);
    dwdwdMdM(l,k) = dwdwdMdM(k,l);
  }
  //std::cout << "dwdM " << dwdM << std::endl;
}

/**
* Get the dD/dMkl
*/
void Lmodel::getdDdMkl(doubleA& dDdMkl,const doubleA& M, int k, int l){
  dDdMkl.resize(M);
  //since it is square
  int i,j,size = M.d0;
  
  for(i=0;i<size;i++) for(j=0;j<size;j++){
    dDdMkl(i,j) = M(k,j)*((i == l)? 1:0) + M(k,i)*((j == l)? 1:0);
  }
}

/**
* Update the distance matrix 
*/
void  Lmodel::update_dist(const doubleA& x, double y, double e_cv, double e){
  doubleA derivative_ok; derivative_ok.resize(n_data);
  uint i;
  for(i=0;i<derivative_ok.d0;i++) 
    derivative_ok(i) = (n_data(i) > 0.1/(1.- lambda(i)))? 1:0;
  double transient_multiplier = 0;
  
  if(! derivative_ok(1)){
    transient_multiplier = 0;
    return;
  }
  
  double w = activation(x);
  doubleA z; computeZs(z,x);
  
  //initialisations
  double Wtmp = lambda(0)*W(0) + w;
  //std::cout<<std::endl<<"e2 "<<e*e<<std::endl;
  //std::cout<<"w "<<w<<std::endl;
  //std::cout<<"p sum_e2 "<<sum_e2<<std::endl;
  sum_e2 = lambda(0)*sum_e2 + w*e*e;
  //std::cout<<"m sum_e2 "<<sum_e2<<std::endl;
  //std::cout<<"e_cv2 "<<e_cv*e_cv<<std::endl;
  //std::cout<<"w "<<w<<std::endl;
  //std::cout<<"p sum_ecv2 "<<sum_ecv2<<std::endl;
  sum_ecv2 = lambda%sum_ecv2 + w*e_cv*e_cv;
  //std::cout<<"m sum_ecv2 "<<sum_ecv2<<std::endl;
  //for numerical stability
  //std::cout<<"sum_e2 "<<sum_e2<<std::endl;
  //std::cout<<"sum_ecv2 "<<sum_ecv2(R-1)<<std::endl;
  transient_multiplier = pow((sum_e2/(sum_ecv2(R-1)+1.e-10)),4);
  if(transient_multiplier>1.0) transient_multiplier=1.0;
  //std::cout<<"t1 "<<(sum_e2/(sum_ecv2(R-1)+1.e-10))<<std::endl;
  //std::cout<<"transient_multiplier "<<transient_multiplier<<std::endl;
  double sum_dJ1dw;
  doubleA dwdM; dwdM.resize(M);
  
  doubleA dJdM;
  getdJdM(dJdM,
    dist, M,
    x, center,
    z, azz, w,
    Wtmp, e_cv, e, gamma, derivative_ok, sum_dJ1dw, dwdM);
  
  if(meta) 
    meta_update(e_cv, Wtmp, w, e, aH, aE, z, azz, derivative_ok, x, center, 
    sum_dJ1dw, dwdM, transient_multiplier, dJdM);
  
  //update the distance metric, use some caution for too large gradients
  double max = absmax(M);
  doubleA delta_m(alpha*dJdM*transient_multiplier);
  intA indices = find(delta_m, 0.1*max, true);
  if(indices.N){
    for(i=0;i<indices.d0;i++){
      int row=indices(i,0);
      int col=indices(i,1);
      alpha(row,col) = alpha(row,col)/2.0;
    }
    std::cout << "Reduced learning rate " << std::endl;
  }
  
  
  M -= alpha*dJdM*transient_multiplier;
  dist = ~M * M;//matmult(~M,M);
  
  update_dist_stat(e_cv, w, z, azz, transient_multiplier, derivative_ok);
}

void Lmodel::meta_update(double e_cv, double W, double w, double e, const doubleA& aH,
			 double aE, const doubleA& z, 
			 const doubleA& azz, const doubleA& derivative_ok,
			 const doubleA& x, const doubleA& center, double dJ1dw,
			 const doubleA& dwdM, double transient_multiplier, const doubleA& dJdM){
  
  doubleA z2(z%z);
  doubleA q(z%derivative_ok/azz);
  doubleA qe(q*e);
  //doubleA tmp(matmult(~z,qe));
  doubleA tmp1(q*scalarProduct(z,qe)); //matmult(q,tmp));
  
  double e_cv2 = e_cv*e_cv;
  double e2 = e*e;
  
  double sum_z = sum((-q*e/W - 2.*tmp1)%aH);
  double sum_y = e_cv2 - 2*sum(e*q%aH);
  double h1 = w*scalarProduct(z,q); //(((doubleA ) matmult(~z,q))(0));
  double dJ1J1dwdw = -e_cv2/(W*W) - (2/W)*sum_z + (2/(W*w))*e2*h1 - (1/(W*W))*sum_y + aE/(W*W*W);
  //std::cout << "first " <<  -e_cv2/(W*W)<< std::endl;
  //std::cout << "second " <<  - (2/W)*sum_z<< std::endl;
  //std::cout << "third " <<  (2/(W*w))*e2*h1<< std::endl;
  //std::cout << "fourth " <<  (1/ << std::endl;
  //std::cout << "fifth " <<  aE/(W*W*W)<< std::endl;
  //std::cout << "dJ1J1dwdw " <<  dJ1J1dwdw<< std::endl;
  doubleA dJ2dJ2dMdM; getdJ2dJ2dMdM(dJ2dJ2dMdM,dist, M, gamma);
  doubleA dwdwdMdM; getdwdwdMdM(dwdwdMdM,M, x, center, w);
  
  doubleA term1(dwdwdMdM*dJ1dw);
  doubleA dwdM2(dwdM*dwdM);
  doubleA term2(dwdM2*dJ1J1dwdw);
  
  doubleA dJJdMdM(term1 + term2 + (w/W)*dJ2dJ2dMdM);
  //std::cout << "dwdwdMdM " << dwdwdMdM << std::endl;
  //std::cout <<"term1 " << term1;
  //std::cout <<"term2 " << term2;
  //std::cout << "dwdM2 " << dwdM2 << std::endl;
  //std::cout << "dJ1J1dwdw " << dJ1J1dwdw << std::endl;
  //std::cout << "(w/W)*dJ2dJ2dMdM" << doubleA ((w/W)*dJ2dJ2dMdM);
  
  double mul = meta_rate * transient_multiplier;
  doubleA aux(mul * dJdM * h);
  doubleA absaux(matabs(aux));
  
  intA ind = find(absaux, 0.1, true);
  //limit the update rate
  uint i;
  for(i=0;i<ind.d0;i++){
    int r = ind(i,0);
    int c = ind(i,1);
    aux(r,c) = 0.1 * ((aux(r,c) > 0) ? 1:-1);
  }
  b = b - aux;
  
  absaux = matabs(aux);
  //prevent numerical overflow
  ind = find(absaux, 10, true);
  for(i=0;i<ind.N;i++){
    intA coord = ind(i);
    int r = coord(0);
    int c = coord(1);
    b(r,c) = 10 * ((b(r,c) > 0) ? 1:-1);
  }
  
  alpha = exp(b);
  
  aux = 1. - (alpha*dJJdMdM) * transient_multiplier;
  //std::cout << "dJJdMdM " << dJJdMdM << std::endl;
  ind = find(aux,0, false);
  for(i=0;i<ind.d0;i++){
    aux(ind(i,0),ind(i,1)) = 0;
  }
  
  h = h*aux - (alpha*dJdM)*transient_multiplier;
  
}

/**
* Finds all the indices of the element greater than the number
*/
intA Lmodel::find(const doubleA& A, double value, bool gt){
  int rsize = A.d0;
  int csize = A.d1;
  
  intA indices;
  
  int i,j;
  for(i=0;i<rsize;i++) for(j=0;j<csize;j++){
    if(gt?(A(i,j) > value):(A(i,j)<value)){
      indices.append(i);
      indices.append(j);
    }
  }

  indices.resize(indices.N/2,2);
  return indices;
}

/**
* Finds the maximum of an Array
*/
double Lmodel::absmax(const doubleA& A){
  int rsize = A.d0;
  int csize = A.d1;
  
  double max = fabs(A(0,0));
  int i,j;
  for(i=0;i<rsize;i++) for(j=0;j<csize;j++)
    max = (max<fabs(A(i,j))) ? fabs(A(i,j)) : max;
    
  return max;
}

void Lmodel::cholesky(doubleA& L,const doubleA& A){
  int i,j,k,size = A.d0;
  L.resize(size,size);
  L = 0;
  
  for(j=0;j<size;j++){
    double d = 0.0;
    for(k=0;k<j;k++){
      double s = 0.0;
      for(i=0;i<k;i++) s += L(k,i)*L(j,i);
      s = (A(j,k) - s)/L(k,k);
      L(j,k) = s;
      d += s*s;
    }
    d = A(j,j) - d;
    L(j,j) = sqrt(MT::MAX(d,0.0));
    for(k=j+1;k<size;k++) L(j,k) = 0.0;
  }
}

void Lmodel::printProj(){
  std::cout << "M : " << std::endl;
  std::cout << M << std::endl;	
  
}

void Lmodel::save(std::ostream& os){
  os <<"RF - parameters: "
    <<initD <<' '
    <<init_lambda <<' '
    <<tau_lambda <<' '
    <<final_lambda <<' '
    <<add_threshold <<' '
    <<meta <<' '
    <<meta_rate
    <<std::endl;

  os<<center <<'\n'
    <<mean_x <<'\n'
    <<mean_y <<'\n'
    <<W <<'\n'
    <<R <<'\n'
    <<lambda <<'\n'
    <<M <<'\n'
    <<n_data <<'\n'
    <<std::endl;

  os<<uprojections <<'\n'
    <<pprojections <<'\n'
    <<sXresYres <<'\n'
    <<betas <<'\n'
    <<MSE <<'\n'
    <<azz <<'\n'
    <<azres <<'\n'
    <<axz <<'\n'
    <<std::endl;

  os<<sum_e2 <<'\n'
    <<sum_ecv2 <<'\n'
    <<aH <<'\n'
    <<aG <<'\n'
    <<aE <<'\n'
    <<gamma <<'\n'
    <<alpha <<'\n'
    <<b <<'\n'
    <<h <<'\n'
    <<trustworthy <<'\n'
    <<dist<<'\n'
    <<std::endl;
}

void Lmodel::load(std::istream& is){
  is>>(const char*)"RF - parameters:"
    >>initD
    >>init_lambda
    >>tau_lambda
    >>final_lambda
    >>add_threshold
    >>meta
    >>meta_rate;

  is>>center
    >>mean_x
    >>mean_y
    >>W
    >>R
    >>lambda
    >>M
    >>n_data;

  is>>uprojections
    >>pprojections
    >>sXresYres
    >>betas
    >>MSE
    >>azz
    >>azres
    >>axz;

  is>>sum_e2
    >>sum_ecv2
    >>aH
    >>aG
    >>aE
    >>gamma
    >>alpha
    >>b
    >>h
    >>trustworthy
    >>dist;
}

/*std::ostream& operator<<(std::ostream& os, const Lmodel& lm){
  //os << "C " << *(lm.center) << std::endl;
  //os << "D " << *(lm.dist) << std::endl;
  os << "M " << lm.M << std::endl;
  os << "n_data " << lm.n_data << std::endl;
  return os;
}*/

void Lmodel::updateApk(const doubleA& x, double w){
  doubleA z; computeZs(z,x);
  doubleA q(z/azz);
  
  double newTerm = w*w*scalarProduct(z,q); //(((doubleA ) matmult(~z,q))(0)); //@todo cf. paper
  apk = lambda(0)*apk + newTerm;
  
  //std::cout << apk << std::endl;
}

double Lmodel::s2pred(const doubleA& x, double w){
  //	double s2 = (*MSE)[R-1] / (W(0) - apk);
  double s2 = MyMSE_R / (W(0) - apk);
  doubleA z; computeZs(z,x);
  doubleA q(z/azz);
  double term = w*scalarProduct(z,q); //(((doubleA ) matmult(~z,q))(0));
  
  return s2 * (1.0+term);
}


//===========================================================================
//
// LWPR itself
//

Lwpr::Lwpr(){
  isInitialized=false;
}

void Lwpr::initialize(uint inDim,uint outDim){
  in_dim=inDim;
  out_dim=outDim;
  //allocate models
  models.resize(out_dim);
  //grab parameters
  w_gen   = MT::Parameter<double>("w_gen",0.2);
  w_prune = MT::Parameter<double>("w_prune",0.9);
  cut_off = MT::Parameter<double>("cut_off",0.001);
  updateD = MT::Parameter<bool>("updateD",true);
  add_proj= MT::Parameter<bool>("add_proj",true);
  w_gen   = MT::Parameter<double>("w_gen",0.2);
  w_prune = MT::Parameter<double>("w_prune",0.9);
  cut_off = MT::Parameter<double>("cut_off",0.001);
  updateD = MT::Parameter<bool>("updateD",true);
  add_proj= MT::Parameter<bool>("add_proj",true);
  rfsno = 0; //number of receptive fields
}

int Lwpr::get_rfs_no(int out_dim){ //number of local models for that output dimension
  return (models(out_dim)).N;
}

double Lwpr::get_proj_average(int out_dim){
  double k=0;
  for(uint x=0;x<models(out_dim).N;x++) 
    k+=models(out_dim)(x).R; //add number of projections for each local model
  return k/get_rfs_no(out_dim);			
}

void Lwpr::map(const doubleA& xUnnorm,doubleA& y) {
  doubleA x;
  normalizeInput(x,xUnnorm);
  if(x.nd==2){ //batch case
    y.resize(x.d0,out_dim);
    for(uint i=0;i<x.d0;i++) for(uint j=0;j<out_dim;j++) y(i,j)=predict(models(j),x[i]);
  }else{ //single datum case
    y.resize(out_dim);
    for(uint i=0;i<out_dim;i++) y(i)=predict(models(i),x); //get prediction for each dimension
  }
}

double Lwpr::predict(const MT::Array<Lmodel>& rfs, const doubleA& x) {
  double wt_sum = 0;
  double sum = 0;
  for(uint i=0;i<rfs.N;i++) { //loop over all models
    Lmodel *lm = &rfs(i); //get local model
    double w = (*lm).activation(x); //get activation at x
    
    if (w<cut_off || !(*lm).trustworthy)
      continue;
    double predy = (*lm).predict(x); //get prediction for x
    wt_sum += w*predy;
    sum += w;
  }
  
  if (sum == 0)
    return 0;
  
  return wt_sum/sum;
}

void Lwpr::learn(const doubleA& xUnnorm, const doubleA& y){
  doubleA x;
  normalizeInput(x,xUnnorm);
  CHECK(x.nd==y.nd,"LWPR can learn batch or single datum");
  if(!isInitialized){
    if(x.nd==2 && y.nd==2) initialize(x.d1,y.d1); //batch of data
    if(x.nd==1 && y.nd==1) initialize(x.d0,y.d0); //single datum
    isInitialized=true;
  }
  uint k,i;
  if(x.nd==2){ //batch case
    CHECK(x.d0==y.d0,"need same number of inputs and outpus in batch");
    for(k=0;k<x.d0;k++) for(i=0;i<out_dim;i++) updaterfs(models(i),x[k],y(k,i));
  }else{ //single datum case
    for(i=0;i<out_dim;i++) updaterfs(models(i),x,y(i));
  }
}

void Lwpr::updaterfs(MT::Array<Lmodel>& rfs, const doubleA& x, double y) {
  if (rfs.N == 0) { 
    rfs.append(Lmodel(x,y)); //at least one model per dimension
    //std::cout << "Lwpr @" << (void *) this << ": Local models: " << ++rfsno << std::endl;
    return;
  }
  //the 3 maximum ws
  doubleA max_ws(3); max_ws=0.;
  //the corresponding indices of the rfs
  intA max_ind(3); max_ind=0;
  uint i;
  for(i=0;i<3;i++){
    max_ws.append(-1);
    max_ind.append(-1);
  }
  
  for(i=0;i<rfs.N;i++) { //loop over all local models
    Lmodel* lm = &rfs(i);
    double w = (*lm).activation(x);
    //don't update if activation less than 0.001
    if (w > 0.001) {
      (*lm).updateMean(x,y);
      double e_cv = (*lm).gete_cv(x,y); //obtain current prediction error = leave-one-out cross validation error of the current point _after_ the regression parameters were updated with the data point
      (*lm).update(x,y);
      double e = (*lm).gete(x,y);
      (*lm).updateError(x,y);
      (*lm).updateError_matlab(e_cv,w);
      if (updateD) (*lm).update_dist(x,y,e_cv,e);
      if ((add_proj) && ((*lm).check_add_projection(x, y)))
	(*lm).add_projection(x, y);
      (*lm).updateStat(w);
      
      (*lm).updateApk(x,w); //With every new point, increment apk.
    }
    double tmp1 = -1;
    int tmp2 = -1;
    int n = i;
    //shift
    for(int j=2;j >= 0;j--) {
      if (w > max_ws(j)) {
	tmp1 = max_ws(j);
	tmp2 = max_ind(j);
	max_ws(j) = w;
	max_ind(j) = n;
	w = tmp1;
	n = tmp2;
      }
    }
    
  }
  
  //if all are less than w_gen then add a new RF
  if (max_ws(2) <= w_gen) {
    Lmodel rf = (rfs(max_ind(2)));
    if (max_ws(2) > 0.1*w_gen && rf.trustworthy)
      rfs.append(Lmodel(rf, x,y));
    else
      rfs.append(Lmodel(x,y));
    //std::cout << "Lwpr @" << (void *) this << ": Local models: " << ++rfsno << std::endl;
  }
  
  //do we need to prune a RF? Prune the one with bigger D
  //equivalent to smaller RF
  if (max_ws(1) > w_prune && max_ws(2) > w_prune) {
    Lmodel rf1 = (rfs(max_ind(1)));
    Lmodel rf2 = (rfs(max_ind(2)));
    doubleA  dist1(rf1.dist);
    doubleA  dist2(rf2.dist);
    int index = (sum(dist1) < sum(dist2)) ? max_ind(2) : max_ind(1); 
    //delete(rfs(index));
    Lmodel* itr = rfs.p;
    int cnt = 0;
    while (itr != rfs.pstop) {
      if (cnt == index)
	rfs.remove(itr-rfs.p);
      itr++;
      cnt++;
    }
    
    std::cout << "Pruned : " << index << std::endl;
  }
  
}

void Lwpr::normalizeInput(doubleA& normIn,const doubleA& unNormIn){
  if(!normMean.N) normIn.referTo(unNormIn);
  else{
    normIn=unNormIn;
    if(normIn.nd==2){ //batch case
      for(uint i=0;i<normIn.d0;i++){
        normIn[i]()-=normMean;
        if(normTrans.nd==1) normIn[i]()*=normTrans;
        if(normTrans.nd==2) normIn[i]()=normTrans*normIn[i];
      }
    }else{
      normIn-=normMean;
      if(normTrans.nd==1) normIn*=normTrans;
      if(normTrans.nd==2) normIn=normTrans*normIn;
    }
  }
}

void Lwpr::save(char *filename){
  std::cout<<"saving model"<<std::endl;
  MT::IOraw=false;
  ofstream os(filename);
  os
    <<"LWPR\n parameters: "
    <<in_dim <<' '
    <<out_dim <<' '
    <<w_gen <<' '
    <<w_prune <<' '
    <<cut_off <<' '
    <<updateD <<' '
    <<add_proj <<std::endl;

  os <<" normalization: " <<normMean <<normTrans <<std::endl;

  uint i,j;
  for(i=0;i<models.N;i++){
    os <<i <<"th output dim, #rfs=" <<models(i).N <<std::endl;
    for(j=0;j<models(i).N;j++) models(i)(j).save(os);
  }
}

void Lwpr::load(char *filename,bool alsoParameters){
  std::cout<<"loading model"<<std::endl;
  MT::IOraw=false;
  isInitialized=true;

  ifstream is(filename);
  is
    >>(const char*)"LWPR" >>(const char*)"parameters:"
    >>in_dim
    >>out_dim
    >>w_gen
    >>w_prune
    >>cut_off
    >>updateD
    >>add_proj;

  if(!alsoParameters){
    w_gen   = MT::Parameter<double>("w_gen",0.2);
    w_prune = MT::Parameter<double>("w_prune",0.9);
    cut_off = MT::Parameter<double>("cut_off",0.001);
    updateD = MT::Parameter<bool>("updateD",true);
    add_proj= MT::Parameter<bool>("add_proj",true);
  }

  is >>(const char*)"normalization:" >>normMean >>normTrans;

  uint i,j,n;
  models.resize(out_dim);
  for(i=0;i<models.N;i++){
    is >> j >>"th output dim, #rfs=" >>n;
    CHECK(i==j,"am I counting wrong?");
    models(i).resize(n);
    for(j=0;j<models(i).N;j++) models(i)(j).load(is);
  }
}

double Lwpr::s2pred(const doubleA& x){
  MT::Array<Lmodel> rfs = models(0); //consider only first output dimension.
  double pred = predict(rfs, x);
  double result = 0, sum_w = 0;
  
  //loop over all local models
  for(uint i=0;i<rfs.N;i++) {
    Lmodel* lm = &rfs(i);
    double w = (*lm).activation(x);
    //don't update if activation less than 0.001
    if (w<cut_off || !(*lm).trustworthy) //@todo should that be included here?
      continue;
    
    double predy = (*lm).predict(x); //prediction of current local model
    
    result += w * ( (pred-predy)*(pred-predy) + (*lm).s2pred(x,w) );
    sum_w += w;
  }
  
  return result/(sum_w*sum_w);
}

void Lwpr::learnNormalization(doubleA& X){
  CHECK(X.nd==2,"normalization should be learned from a batch of data");
  doubleA mean,var,factors;
  mean.resize(X.d1); mean.setZero();
  var.resize(X.d1);  var.setZero();
  factors.resize(X.d1);  factors.setZero();
  uint i;
  for(i=0;i<X.d0;i++){
    const doubleA &x=X[i];
    mean+=x;
    var +=x%x; //this is the element-wise product
  }
  double N=X.d0;
  mean/=(double)N;
  var =(var - (mean % mean)/(double)N)/(double)N;
  for(i=0;i<factors.N;i++) factors(i)=1./::sqrt(var(i));
  useNormalization(mean,factors);
}

